#include "common.h"
#include "sound.h"

typedef struct               
{
 short vol_l,      // Volume left
       vol_r,      // Volume right
       pitch,      // pitch (well , its speed)
       addr,       // addr of wavedate in sound buffer
       ar,         // attack rate
       dr,         // decay rate
       sr,         // sustain rate
       rr,         // release rate
       s_level,    // sustain level
       detailed_pitch; // im not sure with this one 
} voice_data ;

// spu code
int SPUgly_init (void);
int SPUgly_quit(void);
int SPUgly_wait (void);
int SPUgly_mem2sb (void* addr,long sb_addr,long size);
int SPUgly_sb2mem (void* addr,long sb_addr,long size);
int SPUgly_setvoice (u_long num,voice_data *v);
int SPUgly_setkey (short to,long mask);
void SPUgly_enable(void);

struct enc_t {
	int _s_1,_s_2,s_1,s_2;
};

void vag_enc_int(struct enc_t *enc,short *samples, unsigned char *block, int flags);
void vag_enc_init(struct enc_t *enc);

#define	LOOP	1

static int freq,channel,writesize;
static struct enc_t enc[2];
static voice_data voice[2];
static int flags;
static int sb_addr[2] = {0x1000,0x1000+0x40000};
static	unsigned char block[2][16*128]; /* static for DMA */
#define WRITEMAX	0x3f000

int sound_init(int audiofd, layer *info, int stereo) {

    /* Initialize audio device */
	freq = s_freq[info->version][info->sampling_frequency] * 1000;
	channel = stereo;

    return 0;
}

int sound_open() {

	writesize = 0;

	SPUgly_init(); printf (" Spu initialized\n");

	flags = LOOP?6:0;
	vag_enc_init(&enc[0]);
	vag_enc_init(&enc[1]);

    return 0;
}

static void sound_start(void)
{
	int i;
	for(i=0;i<channel;i++) {
		voice_data *v = &voice[i];
		v->vol_l = 127; // maximum
		v->vol_r = 127;                               
		v->ar = 0x80;
		v->sr = 0x7F;
		v->rr = 0x3F;
		v->dr = 0xF;
		v->s_level = 0xF ;
		v->addr = sb_addr[i] ;
		v->pitch = freq*4096/44155 ;
		v->detailed_pitch = 0;
	}
	if (channel==2) {
		voice[0].vol_r=0;
		voice[1].vol_l=0;
	}

	SPUgly_enable();

	for(i=0;i<channel;i++) {
		SPUgly_setvoice (i,&voice[i]); // init voice n
		printf (" Voice [%d] initialized \n",i);
	}

	SPUgly_setkey (0x1,channel==1?1:3); // set voice 0 on
	printf (" Voice started \n");
}

static void sound_finish(void)
{
	printf("sound finish\n");
	memset(block[0],0,64); /* minimum trans is 64 */
	block[0][1] = (flags==2)?3:7; /* end flag */
	SPUgly_mem2sb (block[0],sb_addr[0]+writesize,64);
	if (channel==2)
		SPUgly_mem2sb (block[0],sb_addr[1]+writesize,64);
}

int sound_close(int audiofd) {
	sound_finish();
	sound_start();
//	while(1) {}
	getchar(); /* wait for key input */
    return 0;
}

int psx_sound_write(int audiofd, short *buf0, short *buf1, size_t count)
{
	count/=2;
	if (writesize>=WRITEMAX) {
		//puts("skip");
		return 0;
	}
	//	puts("start");

	{
	//	int endflg = 0;
		int cnt = count/28;
		int j;
		if (count%28) {
			for(j=count%28;j<28;j++) buf0[28*cnt+j]=buf1[28*cnt+j]=0;
			cnt++;
		//	endflg = 1;
		}
		for(j=0;j<cnt;j++) {
            vag_enc_int(&enc[0], buf0 + j * 28, &block[0][j*16],flags);
			if (channel==2) {
	            vag_enc_int(&enc[1], buf1 + j * 28, &block[1][j*16],flags);
	        }
			if (flags==6) flags = 2; //*for loop;
		}
		SPUgly_mem2sb (block[0],sb_addr[0]+writesize,cnt*16);
		if (channel==2)
			SPUgly_mem2sb (block[1],sb_addr[1]+writesize,cnt*16);
		writesize += cnt*16;
		//printf("%d ",writesize);
		if (writesize>=WRITEMAX /* || endflg*/) {
			sound_finish();
			sound_start();
		}
		//printf("%d end\n",cnt*16);
	}

	return count*2;
}

void vag_enc_init(struct enc_t *enc)
{
	enc->s_1 = 0;
	enc->s_2 = 0;
	enc->_s_1 = 0;
	enc->_s_2 = 0;
}

/*
	based on bitmaster's  vagpack
/*

/* 16bit 28 PCM samples to 4bit 16byte XA-ADPCM sample */
void vag_enc_int(struct enc_t *enc,short *samples, unsigned char *block, int flags)
{
static int f[5][2] = {
	{ 0, 0 },
	{  -60, 0 },
	{ -115, 52 },
	{  -98, 55 },
	{ -122, 60 }
};

    int i, j;
    int buffer[5][28];
    int min = (int)0x7fffffff; /* INT_MAX */
    int s_1, s_2;
	int shiftfactor,predictnr;

	/* select form five patterns -> predictnr */
    for ( i = 0; i < 5; i++ ) {
        int max = 0;
        s_1 = enc->_s_1;
        s_2 = enc->_s_2;
        for ( j = 0; j < 28; j ++ ) {
			int ds;
			int s_0 = samples[j];						// s[t-0]
			if (s_0 >  30719) s_0 =  30719;
			if (s_0 < -30720) s_0 = -30720;
            ds = s_0 + (s_1 * f[i][0] + s_2 * f[i][1])/64;
            buffer[i][j] = ds;
            if ( ds< 0 ) ds=-ds;
            if ( ds > max ) max = ds;
            s_2 = s_1;                                  // new s[t-2]
            s_1 = s_0;                                  // new s[t-1]
        }
        
        if ( max < min ) {
            min = max;
            predictnr = i;
        }
        if ( min <= 7 ) {
            predictnr = 0;
            break;
        }
        
    }

// store s[t-2] and s[t-1] in a static variable
// these than used in the next function call

    enc->_s_1 = s_1;
    enc->_s_2 = s_2;
    
//  if ( min > 32767.0 )
//      min = 32767.0;
        
  {
    int shift_mask;
    shift_mask = 0x4000;
    shiftfactor = 0;

	/* select shiftfactor */
    while( shiftfactor < 12 ) {
        if ( shift_mask  & ( min + ( shift_mask >> 3 ) ) )
            break;
        shiftfactor++;
        shift_mask = shift_mask >> 1;
    }
  }

  block[0] = ( predictnr << 4 ) | shiftfactor;
  block[1] = flags;
  {
	int *d_samples = buffer[predictnr];
	short four_bit[28];
	s_1 = enc->s_1;
	s_2 = enc->s_2;

    for ( i = 0; i < 28; i++ ) {
        int s_0 = d_samples[i] + (s_1*f[predictnr][0] + s_2*f[predictnr][1])/64;        int di = ( (s_0<<shiftfactor) + 0x800 ) & 0xfffff000;

        if ( di >  32767 )  di = 32767;
        if ( di < -32768 )  di = -32768;
            
        four_bit[i] = (short) di;

        s_2 = s_1;
        s_1 = (di>>shiftfactor) - s_0;

    }
	enc->s_1 = s_1;
	enc->s_2 = s_2;

    for ( i = 0; i < 28; i += 2 ) {
      block[i/2+2] = ( ( four_bit[i+1] >> 8 ) & 0xf0 ) | ( ( four_bit[i] >> 12 ) & 0xf );
    }
  }
}

// based on  Dodger of CREATURE sample
// Dont forget to visit our page at http://www.in-brb.de/~creature/

/********************************************************/
#define i_reg  (volatile unsigned short*)0x1F801070
#define i_mask (volatile unsigned short*)0x1F801074
#define spu_mvol_l (volatile unsigned short*)0x1F801D80 
#define spu_mvol_r (volatile unsigned short*)0x1F801D82 
#define sp0 (volatile unsigned short*)0x1F801DAA // flag
#define sp1 (volatile unsigned short*)0x1F801DAE // status 
#define spu_keyON_1 (volatile unsigned short*)0x1F801D88 
#define spu_keyON_2 (volatile unsigned short*)0x1F801D8A
#define spu_keyOFF_1 (volatile unsigned short*)0x1F801D8C
#define spu_keyOFF_2 (volatile unsigned short*)0x1F801D8E
#define cd_mvol_l (volatile unsigned short*)0x1F801DB0 
#define cd_mvol_r (volatile unsigned short*)0x1F801DB2 
#define voice_base (volatile unsigned short*)0x1F801C00
#define spu_sb_addr (volatile unsigned short*)0x1F801DA6
#define spu_delay (volatile unsigned short*)0x1F801014 
#define spu_u_1 (volatile unsigned short*)0x1F801D84 // reverberation depth left
#define spu_u_2 (volatile unsigned short*)0x1F801D86 // reverberation depth right
#define spu_u_3 (volatile unsigned short*)0x1F801D98 
#define spu_u_4 (volatile unsigned short*)0x1F801D9A 
#define spu_u_5 (volatile unsigned short*)0x1F801DB4 
#define spu_u_6 (volatile unsigned short*)0x1F801DB6 
#define spu_u_7 (volatile unsigned short*)0x1F801D90
#define spu_u_8 (volatile unsigned short*)0x1F801D92
#define spu_u_9 (volatile unsigned short*)0x1F801D94 
#define spu_u_A (volatile unsigned short*)0x1F801D96
#define spu_u_B (volatile unsigned short*)0x1F801DAC
#define spu_u_D (volatile unsigned short*)0x1F801DA8
#define spu_u_E (volatile unsigned short*)0x1F801DA2
#define spu_dma_1 (volatile long*)0x1F8010C0
#define spu_dma_2 (volatile long*)0x1F8010C4
#define spu_dma_3 (volatile long*)0x1F8010C8
#define d_pcr (volatile long*)0x1F8010F0
#define d_icr (volatile long*)0x1F8010F4
#define t1_mode (volatile long*)0x1F801114
/********************************************************/

/********************************************************/
void SPUgly_enable(void)
{
	*sp0 = 0xC000 ; // Enable playing sound
}
int SPUgly_init (void)
{
	short status,c,*p;

	*d_pcr|=0xB0000;  // enable data transfer through SPU dma channels  
	*spu_mvol_r=0x3FFF; // set main volume  
	*spu_mvol_l=0x3FFF;
	*sp0=0x0; // reset spu
	SPUgly_wait();
	*spu_u_B=0x4;
	*spu_u_1=0x0;  // set reverberation depth
	*spu_u_2=0x0; 
	*spu_keyOFF_1 = 0xFFFF ; // set all voices off 
	*spu_keyOFF_2 = 0xFFFF ; 
	*spu_u_3=0x0; 
	*spu_u_4=0x0;
	*spu_u_7=0x0; 
	*spu_u_8=0x0; 
	*spu_u_9=0x0; 
	*spu_u_A=0x0;
	*cd_mvol_l=0x0; // set cd main volume 
	*cd_mvol_r=0x0;
	*spu_u_5=0x0; 
	*spu_u_6=0x0; 
	status=*sp1 & 0x7FF;
	*spu_sb_addr=0x200;
	SPUgly_wait();
	for (c=0;c<8;c++) *spu_u_D=0x0707;
	SPUgly_wait();
	if (*sp1 & 0x400)
	for (c=0;*sp1 & 0x400;c++)
		if (c>3841) {
			printf ("SPU:T/O ... (wait [WRDY H-> L])");
			return -1;
		}
	SPUgly_wait();
	SPUgly_wait();
	*sp0 = *sp0 & 0xFFCF ; 
	if (status!=*sp1 & 0x7FF)
		for (c=0;status!=*sp1 & 0x7FF;c++)
   if (c>3841)
    {
     printf ("SPU:T/O ... (wait [DMAF CLEAR/W])"); // DMA flag clear wait 
     return -1;
    }

// set voices on off
 *spu_keyON_1 = 0xFFFF ;
 *spu_keyON_2 = 0xFF ;
SPUgly_wait();
SPUgly_wait();
SPUgly_wait();
SPUgly_wait();
 *spu_keyOFF_1 = 0xFFFF ;
 *spu_keyOFF_2 = 0x00FF ;
SPUgly_wait();
SPUgly_wait();
SPUgly_wait();
SPUgly_wait();
 *sp0 = 0x8000 ;// set spu on
 *spu_u_E = 0x1FFF;
}
/********************************************************/
int SPUgly_mem2sb (void* addr,long sb_addr,long size)
{
	long c,i;
//	printf(" %x,%d:",sb_addr,size);

	*spu_sb_addr = sb_addr >> 3 ;

	if (*spu_sb_addr != sb_addr >> 3)
		for (c=0;*spu_sb_addr != sb_addr >> 3;c++)
			if (c>3841*2) /* by bero */
				{printf("SPU fucked up .... WAIT (sb_addr)\n");return 0;}

	*sp0 = *sp0 & 0xFFCF | 0x20 ; // CPU to SPU

	if (*sp0 & 0x30 != 0x20) // if spu not ready to get data
		for (c=0;*sp0 & 0x30 != 0x20;c++) // wait
			if (c>3841) // thats too much
				{printf ("SPU fucked up .... WAIT (receive data)\n");return 0;}

	*spu_delay = *spu_delay & 0xF0FFFFFF | 0x20000000 ;

	*spu_dma_1 = (u_long)addr ;
	*spu_dma_2 = ((size+0x3f)>>6) << 16 | 0x10;
	*spu_dma_3 = 0x01000201 ;       // start CPU->SPU transfer
	while (*spu_dma_3 & 0x01000000) {}; // wait until transfer finished

	return 1;
}
/********************************************************/
int SPUgly_sb2mem (void* addr,long sb_addr,long size)
{
	long c,i;

	*spu_sb_addr =(short) sb_addr >> 3 ; // yo , 8 byte blocks

	if (*spu_sb_addr != sb_addr >> 3)
	for (c=0;*spu_sb_addr != sb_addr >> 3;c++)
		if (c>3841)
			{printf("SPU fucked up .... WAIT (sb_addr)\n");return 0;}

	*sp0 = *sp0 & 0xFFCF | 0x30 ; // SPU to CPU

	if (*sp0 & 0x30 != 0x30) // if spu not ready
		for (c=0;*sp0 & 0x30 != 0x30;c++) // wait 
			if (c>3841) // too much
				{printf ("SPU fucked up .... WAIT (send data)\n");return 0;}

	*spu_delay = *spu_delay & 0xF0FFFFFF | 0x22000000 ;

	*spu_dma_1 = (u_long)addr ;
	*spu_dma_2 = ((size+0x3f)>>6) << 16 ;//| 0x10;
	*spu_dma_3 = 0x01000200 ;       // start SPU->CPU transfer
	while (*spu_dma_3 & 0x01000000) {}; // wait until transfer finished
                                      
	return 1;
}
/********************************************************/
int SPUgly_setvoice (u_long num,voice_data *v)
{
	// dont forget that short is here 2 bytes long (not 1 byte)
	volatile short *b = voice_base + (num*8) ;

	*b       = (v->vol_l << 7)+v->vol_l ;
	*(b+0x1) = (v->vol_r << 7)+v->vol_r ;
	*(b+0x2) = v->pitch ;
	*(b+0x3) = v->addr >> 3 ;
	*(b+0x4) = (v->s_level & 0xF)|((v->dr & 0xF) << 4)|(v->ar << 8) ; 
	*(b+0x5) = (v->rr & 0x3F) | (v->sr << 6) ;
	*(b+0x6) = 0x0 ;
	*(b+0x7) = v->detailed_pitch ; // im not sure 

}
/********************************************************/
int SPUgly_setkey (short to,long mask)
{

	switch (to) {
	case 0 : // OFF
		*spu_keyOFF_1 = mask & 0xFFFF ; 
		*spu_keyOFF_2 = mask >> 16    ; 
		break;
	case 1 : // ON
		*spu_keyON_1 = mask & 0xFFFF ; 
		*spu_keyON_2 = mask >> 16    ;
		break;
	}

	SPUgly_wait();
	SPUgly_wait();
	SPUgly_wait();
	SPUgly_wait();
}
/********************************************************/
int SPUgly_wait (void)
{
	int c;
	long vara=13;

  // wait for some clock cycles
	for (c=0;c<0x60;c++)
		vara = ( (vara << 1)+vara ) << 2 + vara ; // *13

	if (*sp1 & 0x7FF) // if spu isnt ready then wait
		for (c=0;*sp1 & 0x7FF;c++)
			if (c>3841) // thats too long 
				{  printf (" SPUgly_wait :  SPU fucked up .. wait (reset)\n");return 0;}
}
/********************************************************/
int SPUgly_quit(void)
{
/* Prepare_Jump(); */

	*d_icr = *d_icr & 0xDFFFFF | 0x800000 ;

	*sp0 = 0x0 ; // reset (turn spu off)

}

